#include <pic.h>

	__CONFIG(INTIO & WDTEN & UNPROTECT & BOREN & MCLRDIS)

#define LCD

//#define HORIZ_REVERSE 
//#define VERT_REVERSE

#define FRAMES_COUNT

#include "types.h"

//--- PORT DEFINITIONS ---------------------------------------------------
//               76543210
#define  TrisA 0b00001111                
//               ||||||||
//               |||||||ACCELX         
//               ||||||ACCELY        
//               |||||SERVO_IN
//               ||||BUTTON
//               |||SERVO_OUT
//               ||LED
//               |---
//               ---

#ifdef LCD
//               76543210
#define  TrisC 0b00000000
//               ||||||||
//               |||||||D4
//               ||||||D5
//               |||||D6 
//               ||||D7
//               |||E 
//               ||A0
//               |---
//               ---
#else
//               76543210
#define  TrisC 0b00111111
//               ||||||||
//               |||||||TRIM_H
//               ||||||EXT_V
//               |||||EXT_H
//               ||||TRIM_V
//               |||NO_HORIZ
//               ||SHIFT
//               |---
//               ---
#endif

#define	ACCELX    RA0   //        - PIC-13
#define	ACCELY    RA1   //        - PIC-12
#define	SERVO_IN  RA2   //        - PIC-11 (pin of INTerrupt)
#define	BUTTON    RA3   //        - PIC-4 
#define	SERVO_OUT RA4   //        - PIC-3 
#define	LED       RA5   //        - PIC-2 

#ifdef LCD
#define	LCD_A0    RC5   // LCD-4  - PIC-5 
#define	LCD_E     RC4   // LCD-6  - PIC-6 
#define	LCD_D4    RC0   // LCD-11 - PIC-10
#define	LCD_D5    RC1   // LCD-12 - PIC-9
#define	LCD_D6    RC2   // LCD-13 - PIC-8
#define	LCD_D7    RC3   // LCD-14 - PIC-7
#else
#define	TRIM_V    RC3   //        - PIC-7
#define	TRIM_H    RC0   //        - PIC-10
#define	EXT_V     RC1   //        - PIC-9
#define	EXT_H     RC2   //        - PIC-8
#define	NO_HORIZ  RC4   //        - PIC-6
#define	SHIFT     RC5   //        - PIC-5
#endif

#define BUTTON_PRESSED 0
#define SHIFT_ENABLED  0
#define HORIZ_DISABLED 0

#define ACCEL_STABLE 		520
#define MAX_ACCEL_X_VERT	100
#define MAX_ACCEL_Y_HORIZ	50

#define HYSTERESIS_VERT_INNER   20
#define HYSTERESIS_VERT_OUTER   34
#define HYSTERESIS_HORIZ_INNER  30
#define HYSTERESIS_HORIZ_OUTER  43

#define NUM_VOLTAGE_SAMPLES     8  // 2, 4, 8, 16

// Average points;
#define NUM_AVG_SAMPLES 16      // 8, 16, 32=0.5 sec delay

// Maximum input speed to eliminate jitter;
#define MAX_VERT_SERVO_SPEED  10
#define MAX_HORIZ_SERVO_SPEED 10

#ifdef LCD
#include "test.c"
#endif

//--------------------------------------------------------------------------------------
// ISR variables;

#define NUM_SERVO 3

// Servo[0] = Horizontal pan; Servo[1]= Vertical tilt; Servo[2]=AP shutter sw.;
static uint16 Servo[NUM_SERVO];

//--------------------------------------------------------------------------------------
// Filtering;

static bit x_vert_locked;
static bit y_horis_locked;

static bank1 uint16 ServoData_V[NUM_AVG_SAMPLES-1];
static bank2 uint16 ServoData_H[NUM_AVG_SAMPLES-1];

//--------------------------------------------------------------------------------------

static byte CurrentServo;
static bit  ServoStart;
static bit  WaitForGap;

#ifdef FRAMES_COUNT
static uint16 FramesCount;
static uint16 BrokenFramesCount;
#endif

static bit  Shift; // shift servo data to one channel (i.e. 4 -> 5);
static bit  NoHoriz; // horizontal movements servo disabled;

static byte  _Shift; // copy at the begginning of frame;
static byte  _NoHoriz; // copy at the begginning of frame;
static byte _StartCh; // calculates at the begginning of frame;

static uint16 v, h;

//--------------------------------------------------------------------------------------
// Servo related stuff;

#define SERVO_PERIOD	20000
#define SERVO_NEUTRAL	1500
#define SERVO_SPAN	500
#define SERVO_MAX (SERVO_NEUTRAL+SERVO_SPAN)
#define SERVO_MIN (SERVO_NEUTRAL-SERVO_SPAN)
#define SERVO_START	300
#define SERVO_GAP	(3*SERVO_MAX)

#define START_CH	4 // Servo[0] = Rudder(Ch.4); Servo[1]= Switch(Ch.5); Servo[2]=(Ch.6);
// Note if Shift=1, START_CH will be 5 and if NoHoriz=1, it will be 6 without Y (Horiz) impulse;

#define SERVO_HIGH	0 // 0=inverse servo polarity (Hitec Flash 5 needs 0);

//--------------------------------------------------------------------------------------

static uint16 widthHoriz = 0;

void StartWaitForGap(void);

static void interrupt
isr(void)			// Here be interrupt function - the name is unimportant.
{
  // handle the interrupt (insert application code here)

// INTE enabled;
// INTEDG = 1 (0->1);
// needs 3-4 cycles to get into interrupt;

   if (INTF == 1) { // --------- SERVO_EDGE_Interrupt 

     if (WaitForGap) { // found impulse during waiting gap -> need to restart WaitForGap;

#ifdef FRAMES_COUNT
       BrokenFramesCount++;
#endif
       StartWaitForGap();

     } else {

       if (CurrentServo == 0) {
#ifdef FRAMES_COUNT
          FramesCount++;
#endif
          _NoHoriz = (byte) NoHoriz; // copy Y channel enabled state;
          _Shift = (byte) Shift; // copy channel shift;
          _StartCh = START_CH+_Shift+_NoHoriz;
       }

       if (CurrentServo < _StartCh) { // 0, 1, 2, 3, (4)

         // start new synchro;
         SERVO_OUT = SERVO_HIGH;

         T1CON = 0b00000000; // Stop timer;
         TMR1H = ((65536-SERVO_START) >> 8) & 0xFF; 
         TMR1L = (65536-SERVO_START) & 0xFF; // start counting;

         T1CON = 0b00000001; // Resume counting;
         CurrentServo++; 
       }

     } 

     INTF = 0;  /* reset flag */

   } else { // ----------- TMR1_Interrupt 
   
     if (TMR1IF) {

       T1CON = 0b00000000; // Stop timer;

       if (WaitForGap) { // Gap found, because timer finishes count and no INT occurs;
         // Timer1 is inactive - we wait for edge INT;
         WaitForGap = 0;  // now we sit inside gap;
         CurrentServo = 0; // start of sequence;

       } else {

         SERVO_OUT = !SERVO_HIGH;

         if (CurrentServo >= _StartCh) { // 4

          if (CurrentServo == NUM_SERVO+_StartCh-_NoHoriz) { // 2+4=6
            // we finished short "1" impulse followed by last ch., so we will wait the gap;
            StartWaitForGap();

          } else {

            uint16 width;
            if (ServoStart == 1) {

              width = SERVO_START;
              SERVO_OUT = SERVO_HIGH;
              ServoStart = 0;
              CurrentServo++;

            } else {
              // _Shift does not influence array item position; _NoHoriz does;
              // in case of _NoHoriz=1 we will will not out of bounds, because 
              // CurrentServo will have max as NUM_SERVO(3)+START_CH(4)+_Shift-1;
              byte ServoPos = CurrentServo-_StartCh+_NoHoriz;
              width = Servo[ServoPos]; // 4.. -> 0..
              width -= SERVO_START;

              ServoStart = 1; // followed by short "1" for channel delimiter; 

              if (width == 0)
                width = 1;

              if (ServoPos == 0)
                widthHoriz = width;
            }

            width = -width;

            TMR1H = (width >> 8) & 0xFF; ;
            TMR1L = width & 0xFF;

            T1CON = 0b00000001; // Resume counting;
          }

         }

       }

       TMR1IF = 0;         
     }

   }

}

//--------------------------------------------------------------------------------------

void StartWaitForGap(void) {
  WaitForGap = 1; 
  // find 3 slot silence at least in TX sending;
  T1CON = 0b00000000; // Stop timer;
  TMR1H = ((65536-SERVO_GAP) >> 8) & 0xFF; 
  TMR1L = (65536-SERVO_GAP) & 0xFF; // start counting;
  T1CON = 0b00000001; // Resume counting;
}

//--------------------------------------------------------------------------------------

#include "delay.c"

#include "lcd.c"

//--------------------------------------------------------------------------------------

void InitServo(void) {
  // init channels;
  byte i;
  for (i=0; i<NUM_SERVO; i++)
    Servo[i]= SERVO_NEUTRAL;
  for (i=0; i<NUM_AVG_SAMPLES-1; i++) {
    ServoData_V[i]= SERVO_NEUTRAL/2; // Divide by 2 to follow averaging code convention;
    ServoData_H[i]= SERVO_NEUTRAL/2;
  }
  x_vert_locked = 1;
  y_horis_locked = 1;
}

//--- ANALOG SETUP DATA --------------------------------------------------
#define ADChannel_ACCELX 	0
#define ADChannel_ACCELY	1
#define ADChannel_TrimV 	7
#define ADChannel_TrimH		4
#define ADChannel_ExtV 		5
#define ADChannel_ExtH 		6

#define NUM_ADC_Channels	6

#define iAccelX			0
#define iAccelY			1
#define iTrimV			2
#define iTrimH			3
#define iExtV			4
#define iExtH			5

static const byte ADC_Channels[NUM_ADC_Channels] = {
            ADChannel_ACCELX, ADChannel_ACCELY, 
	    ADChannel_TrimV, ADChannel_TrimH,  ADChannel_ExtV, ADChannel_ExtH};
static uint16 ADC_Data[NUM_ADC_Channels];
static bit  button;  

//--------------------------------------------------------------------------------------

uint16 ReadADC(byte piADChannel) {
  byte a = (piADChannel << 2);
  ADCON0 = (ADCON0 & 0b11100011) | a;
  Delay100us(10);
  GODONE = 1;
  while (GODONE) { };

  return (((uint16)ADRESH) << 8) | ADRESL;
}

//--------------------------------------------------------------------------------------

void readChannels(void) {
  byte i;
  uint16 ADC_Val;

  for (i = 0; i < NUM_ADC_Channels; i++) {
    ADC_Data[i] = 0;
  }

  for (i = 0; i < NUM_VOLTAGE_SAMPLES; i++) {
    byte i1;
    for (i1 = 0; i1 < NUM_ADC_Channels; i1++) {
#ifdef LCD
        switch (i1) {
    case iAccelX: case iAccelY:
        ADC_Val = ReadADC(ADC_Channels[i1]); break;
    case iTrimV: 
	ADC_Val = TEST_TrimV; break;
    case iTrimH: 
	ADC_Val = TEST_TrimH; break;
    case iExtV: 
	ADC_Val = TEST_ExtV; break;
    case iExtH: 
	ADC_Val = TEST_ExtH; break;
        }
#else
        ADC_Val = ReadADC(ADC_Channels[i1]);
#endif
      ADC_Data[i1] += ADC_Val;
    }
  }

  for (i = 0; i < NUM_ADC_Channels; i++) {
    ADC_Data[i] /= NUM_VOLTAGE_SAMPLES; // 1:1
  }

  button = BUTTON == BUTTON_PRESSED;
#ifdef LCD
  Shift = (TEST_Shift == SHIFT_ENABLED);
  NoHoriz = (TEST_NoHoriz == HORIZ_DISABLED);
#else
  Shift = (SHIFT == SHIFT_ENABLED);
  NoHoriz = (NO_HORIZ == HORIZ_DISABLED);
#endif
}

//--------------------------------------------------------------------------------------

uint16 long_abs(int16 x) {
 if (x < 0)
   return (uint16) (-x);
 else
   return (uint16) x;
}

//--------------------------------------------------------------------------------------

// Limiting max speed of the value change;
uint16 limit(uint16 x, uint16 x_prev, byte max_speed) {
  int16 delta = x-x_prev;

  if (long_abs(delta) > max_speed) {
    return (delta < 0 ? x-max_speed : x+max_speed);
  } else {
    return x;
  }
}

//--------------------------------------------------------------------------------------

void readAccel(void) {
  int16 v1;
  int16 h1;
  int ext;
  int8 ScaledAccel[2];

#define x_vert (ADC_Data[iAccelX])
#define y_horiz (ADC_Data[iAccelY])

#ifdef LCD
  LCD_CURSOR_LINE_1 (0);  WriteNum (x_vert); // Vertical movement;
  LCD_CURSOR_LINE_1 (5);  WriteNum (y_horiz); // Horizontal movement;
#endif

// Stricking to the middle area;

  if (x_vert_locked == 1) {
	if (x_vert > ACCEL_STABLE-HYSTERESIS_VERT_OUTER)
          x_vert = ACCEL_STABLE;
        else
          x_vert_locked = 0;
  } else {
	if (x_vert > ACCEL_STABLE-HYSTERESIS_VERT_INNER) {
          x_vert = ACCEL_STABLE;		
          x_vert_locked = 1;
        }
  }

  if (y_horis_locked == 1) {
	if (y_horiz > ACCEL_STABLE-HYSTERESIS_HORIZ_OUTER && 
            y_horiz < ACCEL_STABLE+HYSTERESIS_HORIZ_OUTER)
          y_horiz = ACCEL_STABLE;
        else
          y_horis_locked = 0;
  } else {
	if (y_horiz > ACCEL_STABLE-HYSTERESIS_HORIZ_INNER && 
            y_horiz < ACCEL_STABLE+HYSTERESIS_HORIZ_INNER) {
          y_horiz = ACCEL_STABLE;		
          y_horis_locked = 1;
        }
  }

  // Vertical movement;
  x_vert = (ACCEL_STABLE) - x_vert;
  if (x_vert > MAX_ACCEL_X_VERT)
    ScaledAccel[0] = MAX_ACCEL_X_VERT;
  else
    ScaledAccel[0] = (int) x_vert;

  // Horizontal movement;
  if (y_horiz < ACCEL_STABLE) {
    y_horiz = ACCEL_STABLE - y_horiz;
    if (y_horiz > MAX_ACCEL_Y_HORIZ)
      ScaledAccel[1] = -MAX_ACCEL_Y_HORIZ;
    else
      ScaledAccel[1] = -((int) y_horiz);    
  } else
    if (y_horiz > ACCEL_STABLE) {
      y_horiz -= (ACCEL_STABLE);
      if (y_horiz > MAX_ACCEL_Y_HORIZ)
        ScaledAccel[1] = MAX_ACCEL_Y_HORIZ;
      else
        ScaledAccel[1] = (int) y_horiz;    
    } else
      ScaledAccel[1] = 0;
  
  // Vertical movement;
  v1 = ScaledAccel[0]-(MAX_ACCEL_X_VERT/2); // 0..MAX_ACCEL_X_VERT -> -MAX_ACCEL_X_VERT/2..MAX_ACCEL_X_VERT/2; (-50..+50);
  v1 = ((int16) v1)*(SERVO_SPAN/(MAX_ACCEL_X_VERT/2)); // x*500/50=x*10;

  // Horizonal movement;
  h1 = ((int16)ScaledAccel[1])*SERVO_SPAN/MAX_ACCEL_Y_HORIZ; // y*500/50 = y*10;


//--------------------------------------------------------------------------------------
// Changing servo throw by extensions;

#define v_ext (ADC_Data[iExtV])
#define h_ext (ADC_Data[iExtH])

// extension is about %50; radius of 500 around neutral becomes 750; 
// servo will be driven by impulses in range of 750..2250 ms (ext. 0% is default 1000..2000 ms);

  ext = ( long_abs(v1) * ((v_ext+1)/16) ) / 128; // v_ext=0..1023;
  if (v1 < 0) 
    v1 -= (int16) ext;
  else
    v1 += (int16) ext;

  ext = ( long_abs(h1) * ((h_ext+1)/16) ) / 128; // v_ext=0..1023;
  if (h1 < 0) 
    h1 -= (int16) ext;
  else
    h1 += (int16) ext;

//--------------------------------------------------------------------------------------
// Changing servo neutral point by trimming;

#define v_trim (ADC_Data[iTrimV])
#define h_trim (ADC_Data[iTrimH])

// trimming is up to about 50%; 1024/2 ADC value means 0%; 0 means -50%; 1023 means +50%;
// Up to about 256 ms added;

  v1 += ((v_trim+1)/2)-(1024/4);
  h1 += ((h_trim+1)/2)-(1024/4);

//--------------------------------------------------------------------------------------
// performing reversing;

#ifdef HORIZ_REVERSE
  h1 = -h1;
#endif

#ifdef VERT_REVERSE
  v1 = -v1;
#endif

//--------------------------------------------------------------------------------------
// final shifting;

  v = (uint16) (v1+SERVO_NEUTRAL);
  h = (uint16) (h1+SERVO_NEUTRAL);
}

//--------------------------------------------------------------------------------------

// Input: v, h;
// Output: v, h;
// Changes: ServoData_H, ServoData_V;

void doFiltering(void) {
  byte i;
  uint16 unfiltered_v;
  uint16 unfiltered_h;

// averaging filter (sliding window);

  // sum data from NUM_AVG_SAMPLES-1 previous frames and add v, h;
  // shift ServoData arrays towards the begining;
  // preserve unfiltered values for adding to averaging array;

  // Divide by 2 to keep sum in the range, eliminate jitter too.
  v /= 2;
  h /= 2;

  unfiltered_v = v;
  unfiltered_h = h;

  // add first items before, because they will be shifted out
  h += ServoData_H[0];  
  v += ServoData_V[0];  

  for (i=0; i < NUM_AVG_SAMPLES-2; i++) { // 0,1,2
    uint16 t = ServoData_H[i+1];
    ServoData_H[i] = t;
    h += t;
    t = ServoData_V[i+1];
    ServoData_V[i] = t;
    v += t;
  }
  v /= (NUM_AVG_SAMPLES/2); 
  h /= (NUM_AVG_SAMPLES/2);

  // round by 16 to eliminate tiny movements;
//  v.low8 &= 0xFFF0;
//  h.low8 &= 0xFFF0;

  // put latest values in to averaging array;
  ServoData_H[NUM_AVG_SAMPLES-2] = unfiltered_h;
  ServoData_V[NUM_AVG_SAMPLES-2] = unfiltered_v;
}

//--------------------------------------------------------------------------------------
// speed limiting;
void speedLimiting(void) {
  v = limit(v, ServoData_V[NUM_AVG_SAMPLES-2], MAX_VERT_SERVO_SPEED);
  h = limit(h, ServoData_V[NUM_AVG_SAMPLES-2], MAX_HORIZ_SERVO_SPEED);
}
                                                                                                      
// Find if channels Cell1, Cell2, Cell3 and Temp are available;
void process(void) {

  readChannels();

  readAccel();

  speedLimiting();

  doFiltering();

//--------------------------------------------------------------------------------------
// update servo movements;

  GIE = 0; Servo[1] = v; GIE = 1;
  GIE = 0; Servo[0] = h; GIE = 1;
  if (button) {
    GIE = 0; Servo[2] = SERVO_MIN; GIE = 1;
  } else {
    GIE = 0; Servo[2] = SERVO_MAX; GIE = 1;
  }  

//0123456789012345
//xxxx yyyy   ffff
//ssss ssss   ffff

#ifdef LCD
  LCD_CURSOR_LINE_2 (0);  WriteNum (Servo[0]);
  LCD_CURSOR_LINE_2 (5);  WriteNum (Servo[1]);
//  LCD_CURSOR_LINE_2 (8);  WriteNum (v);
//  LCD_CURSOR_LINE_2 (12);  WriteNum (h);  
#ifdef FRAMES_COUNT
  LCD_CURSOR_LINE_1 (12);  WriteNum (FramesCount);
//  LCD_CURSOR_LINE_2 (12);  WriteNum (BrokenFramesCount);  
#endif
//  LCD_CURSOR_LINE_1 (10);  Write1Digit (_StartCh);  
  LCD_CURSOR_LINE_2 (12);  WriteNum (widthHoriz);  
#endif
  LED = button;
}

//--------------------------------------------------------------------------------------

void main(void) {

  TRISA = TrisA;
  TRISC = TrisC;
  OPTION = 0b11001110; // no pull-up, int/RA2 on 0->1; Prescalder 64 -> ~1 sec. to WDT;
  PORTA = 0; 
  PORTC = 0;
  button = 0;

  CMCON0 = 0b00000111; // comparator is turned off;

  ANSEL = (1 << ADChannel_ACCELX) | (1 << ADChannel_ACCELY) |
#ifdef LCD
          0;
#else
          (1 << ADChannel_TrimV) | (1 << ADChannel_TrimH) | 
          (1 << ADChannel_ExtV ) | (1 << ADChannel_ExtH );
#endif


  ADCON1 = 0b01010000;      //  Select the Clock as Fosc/16

  ADIE = 0;

  ADCON0 = 0b10000001;     //  Turn on the ADC
                              //   Bit 7 - Right Justified Sample
                              //   Bit 6 - Dont Use VRef
                              //   Bit 4:2 - Channel 0
                              //   Bit 1 - Do not Start
                              //   Bit 0 - Turn on ADC
  LED=1;

#ifdef LCD
  LCD_Init();          
#endif
   
  InitServo();
  CurrentServo = 0;
  ServoStart = 0;
  SERVO_OUT = !SERVO_HIGH;
#ifdef FRAMES_COUNT
  BrokenFramesCount = 0;
  FramesCount = 0;
#endif

  // Init Timer1;
  TMR1IF = 0;           
  INTF = 0;

  StartWaitForGap();

  PIE1 = 1; 

  PEIE = 1;
  INTE = 1;
  GIE = 1;

  process();

  Delay10ms(20);
  LED=0;
  Delay10ms(20);
  LED=1;
  Delay10ms(20);
  LED=0;

  while (1) {
    CLRWDT(); 
    process();

  };
}

